from pydantic import BaseModel, model_validator, Field, AliasChoices, ConfigDict
from typing_extensions import Self, Literal, Optional
from typing import Union
from GANDLF.configuration.validators import validate_class_list, validate_norm_type
from GANDLF.models import global_models_dict

# Define model architecture options
ARCHITECTURE_OPTIONS = Literal[tuple(global_models_dict.keys())]
# Define model norm_type options
NORM_TYPE_OPTIONS = Literal["batch", "instance", "none"]
# Define model final_layer options
FINAL_LAYER_OPTIONS = Literal[
    "sigmoid",
    "softmax",
    "logsoftmax",
    "tanh",
    "identity",
    "logits",
    "regression",
    "None",
    "none",
]
TYPE_OPTIONS = Literal["torch", "openvino"]
DIMENSIONS_OPTIONS = Literal[2, 3]


# You can define new parameters for model here. Please read the pydantic documentation.
# It allows extra fields in model dict.
class ModelConfig(BaseModel):
    model_config = ConfigDict(
        extra="allow"
    )  #  it allows extra fields in the model dict
    dimension: Optional[DIMENSIONS_OPTIONS] = Field(
        description="model input dimension (2D or 3D)."
    )
    architecture: ARCHITECTURE_OPTIONS = Field(description="Architecture.")
    final_layer: FINAL_LAYER_OPTIONS = Field(description="Final layer.")
    norm_type: Optional[NORM_TYPE_OPTIONS] = Field(
        description="Normalization type.", default="batch"
    )  # TODO: check it again
    base_filters: Optional[int] = Field(
        description="Base filters.", default=None, validate_default=True
    )  # default is 32
    class_list: Union[list, str] = Field(default=[], description="Class list.")
    num_channels: Optional[int] = Field(
        description="Number of channels.",
        validation_alias=AliasChoices(
            "num_channels", "n_channels", "channels", "model_channels"
        ),
        default=3,
    )  # TODO: check it
    type: TYPE_OPTIONS = Field(description="Type of model.", default="torch")
    data_type: str = Field(description="Data type.", default="FP32")
    save_at_every_epoch: bool = Field(default=False, description="Save at every epoch.")
    amp: bool = Field(default=False, description="Automatic mixed precision")
    ignore_label_validation: Union[int, None] = Field(
        default=None, description="Ignore label validation."
    )  # TODO:  To check it
    print_summary: bool = Field(default=True, description="Print summary.")

    @model_validator(mode="after")
    def model_validate(self) -> Self:
        # TODO: Change the print to logging.warnings
        self.class_list = validate_class_list(
            self.class_list
        )  # init and validate the class_list parameter
        self.norm_type = validate_norm_type(
            self.norm_type, self.architecture
        )  # init and validate the norm type
        if self.amp is False:
            print("NOT using Mixed Precision Training")

        if self.save_at_every_epoch:
            print(
                "WARNING: 'save_at_every_epoch' will result in TREMENDOUS storage usage; use at your own risk."
            )  # TODO: It is better to use logging.warning

        if self.base_filters is None:
            self.base_filters = 32
            print("Using default 'base_filters' in 'model': ", self.base_filters)

        return self



import logging
from typing import Optional, Union


from typing import Type
from pydantic import BaseModel, ValidationError, create_model
from pydantic_core import ErrorDetails


def initialize_key(
    parameters: dict, key: str, value: Optional[Union[str, float, list, dict]] = None
) -> dict:
    """
    This function initializes a key in the parameters dictionary to a value if it is absent.

    Args:
        parameters (dict): The parameter dictionary.
        key (str): The key to initialize.
        value (Optional[Union[str, float, list, dict]], optional): The value to initialize. Defaults to None.

    Returns:
        dict: The parameter dictionary.
    """
    if parameters is None:
        parameters = {}
    if key in parameters:
        if parameters[key] is not None:
            if isinstance(parameters[key], dict):
                # if key is present but not defined
                if len(parameters[key]) == 0:
                    parameters[key] = value
    else:
        parameters[key] = value  # if key is absent

    return parameters


# Define custom error messages. The key must be a pydantic type error.
CUSTOM_MESSAGES = {
    "literal_error": "The input must be a valid option, please read the documentation",
    "missing": "This parameter is required. Please define it",
}


def convert_errors(e: ValidationError, custom_messages=None) -> list[ErrorDetails]:
    if custom_messages is None:
        custom_messages = CUSTOM_MESSAGES
    new_errors: list[ErrorDetails] = []
    for error in e.errors():
        custom_message = custom_messages.get(error["type"])
        if custom_message:
            ctx = error.get("ctx")
            error["msg"] = custom_message.format(**ctx) if ctx else custom_message
        new_errors.append(error)
    return new_errors


def extract_messages(errors: list[ErrorDetails]) -> list[str]:
    error_messages: list[str] = []
    for error in errors:
        location = error.get("loc")
        if len(location) == 2:
            message = f"Configuration Error: Parameter: ({location[0]}, {location[1]}) - {error['msg']}"
        else:
            message = (
                f"Configuration Error: Parameter: ({location[0]}) - {error['msg']}"
            )
        error_messages.append(message)
    return error_messages


def handle_configuration_errors(e: ValidationError):
    messages = extract_messages(convert_errors(e))
    for message in messages:
        logging.error(message)


def combine_models(base_model: Type[BaseModel], extra_model: Type[BaseModel]):
    """Combine base model with an extra model dynamically."""
    fields = {}
    # Collect base model fields
    for field_name, field_info in base_model.model_fields.items():
        fields[field_name] = (
            field_info.annotation,
            field_info.default if field_info.default is not Ellipsis else ...,
        )

    # Add fields from the extra model
    for field_name, field_info in extra_model.model_fields.items():
        fields[field_name] = (
            field_info.annotation,
            field_info.default if field_info.default is not Ellipsis else ...,
        )

    # Return the new dynamically combined model
    return create_model(base_model.__name__, **fields)



import torch.nn as nn
from GANDLF.models.seg_modules.Interpolate import Interpolate


class UpsamplingModule(nn.Module):
    def __init__(
        self,
        input_channels,
        output_channels,
        conv=nn.Conv2d,
        conv_kwargs=None,
        scale_factor=2,
        interpolation_mode="trilinear",
    ):
        """
        A module for upsampling the input tensor using interpolation and convolution

        Args:
            input_channels (int): Number of input channels
            output_channels (int): Number of output channels
            conv (nn.Module): Convolutional layer type to use
            conv_kwargs (dict): Optional dictionary of keyword arguments to pass to the convolutional layer
            scale_factor (float or tuple of float): Upsampling factor for the interpolation
            interpolation_mode (str): Interpolation mode for the upsampled tensor

        """
        nn.Module.__init__(self)
        if conv_kwargs is None:
            conv_kwargs = {"kernel_size": 1, "stride": 1, "padding": 0, "bias": True}

        self.interp_kwargs = {
            "size": None,
            "scale_factor": scale_factor,
            "mode": interpolation_mode,
            "align_corners": True,
        }
        self.interpolate = Interpolate(interp_kwargs=self.interp_kwargs)

        self.conv0 = conv(input_channels, output_channels, **conv_kwargs)

    def forward(self, x):
        """
        Passes the input tensor through the interpolation and convolutional layers and returns the output tensor.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            torch.Tensor: The output tensor.
        """
        x = self.conv0(self.interpolate(x))
        return x



from typing import Callable


def parse_opacus_params(params: dict, initialize_key: Callable) -> dict:
    """
    Function to set defaults and augment the parameters related to making a trained model differentially
    private with respect to the training data.

    Args:
        params (dict): Training parameters.
        initialize_key (Callable): Function to fill in value for a missing key.

    Returns:
        dict: Updated training parameters.
    """

    if not isinstance(params["differential_privacy"], dict):
        print(
            "WARNING: Non dictionary value for the key: 'differential_privacy' was used, replacing with default valued dictionary."
        )
        params["differential_privacy"] = {}
    # these are some defaults
    params["differential_privacy"] = initialize_key(
        params["differential_privacy"], "noise_multiplier", 10.0
    )
    params["differential_privacy"] = initialize_key(
        params["differential_privacy"], "max_grad_norm", 1.0
    )
    params["differential_privacy"] = initialize_key(
        params["differential_privacy"], "accountant", "rdp"
    )
    params["differential_privacy"] = initialize_key(
        params["differential_privacy"], "secure_mode", False
    )
    params["differential_privacy"] = initialize_key(
        params["differential_privacy"], "allow_opacus_model_fix", True
    )
    params["differential_privacy"] = initialize_key(
        params["differential_privacy"], "delta", 1e-5
    )
    params["differential_privacy"] = initialize_key(
        params["differential_privacy"], "physical_batch_size", params["batch_size"]
    )

    if params["differential_privacy"]["physical_batch_size"] > params["batch_size"]:
        print(
            f"WARNING: The physical batch size {params['differential_privacy']['physical_batch_size']} is greater"
            f"than the batch size {params['batch_size']}, setting the physical batch size to the batch size."
        )
    params["differential_privacy"]["physical_batch_size"] = params["batch_size"]

    # these keys need to be parsed as floats, not strings
    for key in ["noise_multiplier", "max_grad_norm", "delta", "epsilon"]:
        if key in params["differential_privacy"]:
            params["differential_privacy"][key] = float(
                params["differential_privacy"][key]
            )

    return params
